/*=========================================================================

  Program:   gapfilling
  Language:  C++

  Copyright (c) Jordi Inglada. All rights reserved.

  See LICENSE for details.

  This software is distributed WITHOUT ANY WARRANTY; without even
  the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
  PURPOSE.  See the above copyright notices for more information.

=========================================================================*/

#include "otbWrapperApplication.h"
#include "otbWrapperApplicationFactory.h"
#include "otbTemporalGapFilling.h"

namespace otb
{
namespace Wrapper
{

class ImageTimeSeriesGapFilling: public Application
{
public:
  typedef ImageTimeSeriesGapFilling Self;
  typedef Application                   Superclass;
  typedef itk::SmartPointer<Self> Pointer; 
  typedef itk::SmartPointer<const Self> ConstPointer;

  itkNewMacro(Self);
  itkTypeMacro(ImageTimeSeriesGapFilling, otb::Application);

  using ImageType = otb::VectorImage<float, 2>;
  using LinearFunctorType =
    GapFilling::LinearGapFillingFunctor<ImageType::PixelType>;
  using SplineFunctorType =
    GapFilling::SplineGapFillingFunctor<ImageType::PixelType>;

private:
  void DoInit()
  {
    SetName("ImageTimeSeriesGapFilling");
    SetDescription("Time series gapfilling.");

    // Documentation
    SetDocName("Image Time Series Gap Filling");
    SetDocLongDescription("This application performs .");
    SetDocLimitations("None");
    SetDocAuthors("Jordi Inglada");

    AddDocTag(Tags::Filter);
    AddDocTag("MultiTemporal");
    
    AddParameter(ParameterType_String, "in", 
                 "Input time series");
    AddParameter(ParameterType_String, "mask", 
                 "Mask time series");

    AddParameter(ParameterType_String, "out", 
                 "Output time series");
    AddParameter(ParameterType_Int, "comp", "Number of components per date.");
    AddParameter(ParameterType_String, "it", 
                 "Interpolation type (linear, spline)");
    AddParameter(ParameterType_String, "id", 
                 "Input date file");
    MandatoryOff("id");
    AddParameter(ParameterType_String, "od", 
                 "Output date file");
    MandatoryOff("od");
  }

  void DoUpdateParameters()
  {
  }

  void DoExecute()
  {  
    auto cpd = GetParameterInt("comp");
    std::string in_date_file{""};
    std::string out_date_file{""};
    if(IsParameterEnabled("id"))
      in_date_file = GetParameterString("id");
    if(IsParameterEnabled("id") && IsParameterEnabled("od"))
      out_date_file = GetParameterString("od");
    if(GetParameterString("it") == "linear")
      {
      std::cout << "Using linear interpolation and " << cpd
                << " components per date " << std::endl;
      GapFilling::gapfill_time_series<ImageType, 
                                      LinearFunctorType>(
                                        GetParameterString("in"),
                                        GetParameterString("mask"),
                                        GetParameterString("out"),
                                        cpd,
                                        in_date_file,
                                        out_date_file);
      }
    else
      {
      std::cout << "Using spline interpolation and " << cpd
                << " components per date " << std::endl;
      GapFilling::gapfill_time_series<ImageType, 
                                      SplineFunctorType>(
                                        GetParameterString("in"),
                                        GetParameterString("mask"),
                                        GetParameterString("out"),
                                        cpd,
                                        in_date_file,
                                        out_date_file);
      }
  }
};
} // end namespace Wrapper
} // end namespace otb

OTB_APPLICATION_EXPORT(otb::Wrapper::ImageTimeSeriesGapFilling)
